/* RDATE.C - Copyright (C) 1992,1993 MurkWorks, Inc. All Rights Reserved

   Brad K. Clements, 11/5/93

   Implement RDATE tcp/udp client to set time on Novell server

*/

#include <stdio.h>
#include <errno.h>
#include <niterror.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <dos.h>
#include <io.h>
#include <advanced.h>
#include <process.h>
#include <string.h>
#include <signal.h>
#include <nwconn.h>
#include <nwbindry.h>
#include <nwdir.h>
#include <nwcntask.h>
#include <ctype.h>
#include <netdb.h>
#include <nwserial.h>
#include <library.h>
#include <nwmisc.h>
#include <nwipxspx.h>

#include <stddef.h>
#include <io.h>
#include <stdarg.h>

#define	MODE_TCP	1
#define	MODE_UDP	2
#define	MODE_POLL	4
#define	MODE_QUIET	8

#define	TIME_RETRIES	5

#define	IPPORT_RDATE	37

int	variance;
int	maxchange = 3660;

NETDB_DEFINE_CONTEXT

void	
CFG_Report_Error(char *s, ...)
{
	va_list	arg;
	char		buff[256];

	va_start(arg,s);
	vsprintf(buff,s,arg);
	va_end(arg);
	ConsolePrintf("%s\n\r",buff);
}


void
xperror(char *s)
{
	CFG_Report_Error("RDATE Error:%s:%s",s ? s : "", strerror(errno));
}


void
Settime(long now, int mode)
{
	time_t	mynow, delta;
	struct	tm	*lt;
	int	rc;
	char	buff[64];

	time(&mynow);
	delta = abs(mynow - now);
	lt = localtime(&now);


	strftime(buff,63,"%X %x",lt);
	
	if(mode & MODE_POLL) {
		ConsolePrintf("Remote server time is %s, a difference of %d seconds\r\n",buff, delta);
		return;
	}
	if(variance && (delta <= variance))
		return;

	if(maxchange && delta > maxchange) {
	 	ConsolePrintf("RDATE: Obsurb time change of %d seconds ignored because it exceeds maxchange\r\n",
			delta);
		return;
	}

	rc = SetFileServerDateAndTime(lt->tm_year,lt->tm_mon + 1, lt->tm_mday,
		lt->tm_hour,lt->tm_min, lt->tm_sec);
	if(rc)
		xperror("SetFileServerDateAndTime");
	else if(!(mode & MODE_QUIET))
		ConsolePrintf("RDATE:Server time changed to %s, %d seconds change\r\n",buff, delta);
}

int
GetRDate(struct sockaddr_in server, int mode, char *hostname)
{
	int			fd, rc;
	struct sockaddr_in	me;
	fd_set			readfds;
	struct	timeval	t;
	long			now, then, delta;
	int			tries = 0;

	if(mode & MODE_UDP)  {
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		if(fd < 0) {
			xperror("socket_u");
			return(-1);
		}

		memset(&me, 0, sizeof(me));
		me.sin_port = 0; /* ntohs(900); */
		server.sin_port = htons(IPPORT_RDATE);
		if(bind(fd, (struct sockaddr *) &me, sizeof(me)) < 0) {
			xperror("bind_u");
			close(fd);
			return(-2);
		}
		if(!(mode & MODE_QUIET))
			ConsolePrintf("RDATE:Udp Xmit ");
		while(++tries < TIME_RETRIES) {
			FD_ZERO(&readfds);
			FD_SET(fd, &readfds);
			t.tv_sec = 5;
			t.tv_usec = 0;
			then = GetCurrentTicks();
			sendto(fd, (char *) &now,sizeof(long), 0, (struct sockaddr *) &server, sizeof(server));
			if(!(mode & MODE_QUIET))
				ConsolePrintf("X ");

			rc = select(FD_SETSIZE, &readfds, NULL,NULL, &t);
			if(rc < 1) {
				continue;
			}
			rc = sizeof(me);
			if(recvfrom(fd, (char *) &now, sizeof(now), 0, (struct sockaddr *) &me, &rc) < 0) {
				xperror("recvfrom");
				close(fd);
				return(-4);
			}
			else
				break;
		}
		close(fd);
		if(tries >= TIME_RETRIES) {
			if(!(mode & MODE_QUIET))
				ConsolePrintf("\r\nRDATE:UDP Timed out on host: %s\r\n", hostname);
			return(-3);
		} 
		else
			if(!(mode & MODE_QUIET))
				ConsolePrintf("ok\r\n");

		now = ntohl(now);
		now -= 2208988800;		/* convert to 1970 time */
		delta = GetCurrentTicks() - then;		/* delta has sec's gone by */
		now += delta/36;		/* ticks to seconds, divided by 2 */
		Settime(now, mode);
		return(0);
	}
	else {
		int	iread;

		fd = socket(AF_INET, SOCK_STREAM, 0);
		if(fd < 0) {
			xperror("socket_t");
			return(-1);
		}

		memset(&me, 0, sizeof(me));
		server.sin_port = htons(IPPORT_RDATE);
		if(bind(fd, (struct sockaddr *) &me, sizeof(me)) < 0) {
			xperror("bind_t");
			close(fd);
			return(-2);
		}
	
		if(connect(fd, (struct sockaddr *) &server, sizeof(server)) < 0) {
			if(!(mode & MODE_QUIET))
				ConsolePrintf("\r\nRDate TCP Timed out on host: %s\r\n", hostname);

			xperror("connect_t");
			close(fd);
			return(-3);
		}

		then = GetCurrentTicks();
		iread = 0;
		while(iread < sizeof(now)) {
			int rc = read(fd, ((char *) &now) + iread, sizeof(now)-iread);
			if(rc < 1) {
			 	ConsolePrintf("\r\nRDate read error %d\r\n",rc);
				xperror("read_t");
				close(fd);
				return(-4);
			}
			iread += rc;
		}
		now = ntohl(now);
		now -= 2208988800;		/* convert to 1970 time */
		delta = GetCurrentTicks() - then;		/* delta has ticks gone by */
		now += delta/36;		/* ticks to seconds, divided by 2 */
		Settime(now, mode);
		close(fd);
		return(0);
	}
}


void
Usage() 
{
	ConsolePrintf("\r\nUsage:\r\n   rdate [/u] [/c] [/p minutes] [/v secs] [/q] [/m secs] server_name ...\r\n");
	ConsolePrintf("   remote_server_name -- the IP address or name of the remote server\r\n");
	ConsolePrintf("            /u        -- use UDP instead of TCP\r\n");
	ConsolePrintf("            /c        -- just check the time and print the difference\r\n");
	ConsolePrintf("            /p nn     -- Continually poll and set the time every nn minutes\r\n");
	ConsolePrintf("            /v zz     -- Don't change the time if the variance is less\r\n");
	ConsolePrintf("                         than zz seconds\r\n");
	ConsolePrintf("            /q        -- Don't report when the time is changed\r\n");
	ConsolePrintf("            /m zz     -- Don't change the time if the delta is greater than zz seconds\r\n");
#ifdef	OVER
	ConsolePrintf("            /o nn     -- Don't use the timezone offset from the set timezone\r\n");
	ConsolePrintf("                         command. Instead use an offset of nn minutes from GMT\r\n");
#endif
}


int
main(int argc, char *argv[])
{
	int	mode = MODE_TCP;
	int	polltime = 0;
	int	x;
	struct	sockaddr_in	so;
	char	*hostname;
	int	firsthost = 0;
	long	offset = 0;
	int	hostptr;
	int	hostcount;
	int	lastfailed = 1;

	if(argc < 2) {
		Usage();
		return(0);
	}

	tzset();

	for(x=1; x < argc; x++) {
	 	if(!stricmp(argv[x],"/c")) {
			mode |= MODE_POLL;
		}
		else if(!stricmp(argv[x],"/q")) {
			mode |= MODE_QUIET;
		}
		else if(!stricmp(argv[x],"/u")) {
			mode &= ~MODE_TCP;
			mode |= MODE_UDP;
		}
		else if(!stricmp(argv[x],"/p")) {
		 	if(x+1 == argc) {
				ConsolePrintf("/p option missing argument\r\n");
				return(0);
			}
			polltime = atoi(argv[++x]);
		}
		else if(!stricmp(argv[x],"/v")) {
		 	if(x+1 == argc) {
				ConsolePrintf("/v option missing argument\r\n");
				return(0);
			}
			variance = atoi(argv[++x]);
		}
		else if(!stricmp(argv[x],"/m")) {
		 	if(x+1 == argc) {
				ConsolePrintf("/m option missing argument\r\n");
				return(0);
			}
			maxchange = atoi(argv[++x]);
		}

#ifdef	OVER
		else if(!stricmp(argv[x],"/o")) {
			if(x+1 == argc) {
				ConsolePrintf("/o option missing argument\r\n");
				return(0);
			}
			offset = atoi(argv[++x]);
		}
#endif
		else if(*argv[x] == '/') {
		 	ConsolePrintf("Unknown option %s\r\n",argv[x]);
			Usage();
			return(0);
		}
		else {
			firsthost = x;
			break;
		}
	}
	if(!firsthost) {
		ConsolePrintf("No remote host name specified\r\n");
		return(0);
	}

	hostptr = firsthost;
	hostcount = argc - firsthost;

	if(polltime)
		ConsolePrintf("RDATE:Will poll server every %d minutes\r\n",polltime);

	while(1) {
		hostname = argv[hostptr];
		if(!(mode & MODE_QUIET) && lastfailed)
			ConsolePrintf("RDATE:Trying Host %s\r\n", hostname);
	
		memset(&so, 0, sizeof(so));
		so.sin_addr.s_addr = inet_addr(hostname);
		so.sin_family = AF_INET;
	
		if (so.sin_addr.s_addr == -1) {
			struct hostent *h;
	
			h = gethostbyname(hostname);
			if(!h) {
			 	xperror("gethostbyname");
				ConsolePrintf("RDATE:Unkown host name, can't resolve: %s\r\n", hostname);
				return(0);
			}
			memcpy(&so.sin_addr.s_addr, h->h_addr, sizeof(so.sin_addr.s_addr));
		}
		x = GetRDate(so, mode, hostname);
		if(x) {
			lastfailed = 1;
		 	hostptr++;
			if(hostptr == argc) {
				if(!polltime) {
					ConsolePrintf("RDATE:All hosts failed\r\n");
					return(0);
				}
				hostptr = firsthost;
			}
			hostcount--;
			if(!hostcount) {
			 	ConsolePrintf("RDATE:All hosts failed, retrying in %d minutes\r\n", polltime);
				goto cont;
			}
			continue;
		}
		else
			lastfailed = 0;
		if(!polltime)
			break;
cont:;
	 	delay(polltime * 1000 * 60);
		hostcount = argc - firsthost;
	}
	return(0);
}




